/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：28所 C++ 信息安全性设计准则
 * 项目描述：用于检查C++源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package com.keyware.sonar.cxx.rules.checkers;

import com.keyware.sonar.cxx.SubscriptionAstVisitor;
import com.sonar.cxx.sslr.api.AstNode;
import com.sonar.cxx.sslr.api.AstNodeType;
import com.sonar.cxx.sslr.api.Grammar;
import org.sonar.check.Priority;
import org.sonar.check.Rule;
import org.sonar.cxx.parser.CxxGrammarImpl;
import org.sonar.cxx.squidbridge.annotations.ActivatedByDefault;
import org.sonar.cxx.squidbridge.annotations.SqaleConstantRemediation;
import org.sonar.cxx.squidbridge.checks.SquidCheck;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 使用关键资源时指定资源所在的路径
 * 1.获取到声明预计
 * 2.判断是不是路径
 * 3.判断是不是绝对路径
 *
 * @author RenFengJiang
 * @date 2024/1/22
 */
@Rule(key = "PathVerifyChecker", name = "使用关键资源时指定资源所在的路径", description = "使用关键资源时指定资源所在的路径", priority = Priority.INFO, tags = {"28suo"})
@ActivatedByDefault
@SqaleConstantRemediation("5min")
public class PathVerifyChecker extends SquidCheck<Grammar> {
    @Override
    public void init() {
        // 指定当前访问器需要访问的节点类型，functionBody（函数）主体节点
        this.subscribeTo(
                CxxGrammarImpl.functionBody
        );
    }

    /**
     * 访问AST节点
     *
     * @param node 要处理的AST节点，该节点类型为通过subscribeTo方法订阅的类型
     */
    @Override
    public void visitNode(@Nonnull AstNode node) {
        IfBuffer ifBuffer = new IfBuffer(this);
        ifBuffer.accept(node);

    }

    //判断是不是路径
    public static boolean isPath(String str) {
        // 检查字符串是否至少包含一个有效的路径分隔符
        return (str.contains("/") || str.contains("\\")); // 注意：Java中需要双反斜杠来表示单个反斜杠字符
    }

    //windows下判断是否是绝对路径
    public static boolean isWindowsAbsolutePath(String path) {
        // 正则表达式匹配以驱动器号（如C:）开始的路径
        String regex = "^([A-Za-z]):\\\\.*";
        Pattern pattern = Pattern.compile(regex);
        Matcher matcher = pattern.matcher(path);
        return matcher.matches();
    }

    //判断linux下判断是否是绝对路径
    public static boolean isLinuxAbsolutePath(String path) {
        // 正则表达式匹配以根目录（/）开始的路径
        String regex = "^/";
        Pattern pattern = Pattern.compile(regex);
        Matcher matcher = pattern.matcher(path);

        return matcher.find(); // 使用find()而不是matches()，因为matches()要求整个字符串都符合模式
    }

    class IfBuffer extends SubscriptionAstVisitor {

        List lists = new ArrayList<String>();

        public IfBuffer(SquidCheck<Grammar> checker) {
            super(checker);
        }

        @Override
        public List<AstNodeType> visitNodeTypes() {
            // 指定当前访问器需要访问的节点类型，这里指定了simpleDeclaration（简单声明）节点类型
            return List.of(CxxGrammarImpl.functionBody);
        }

        @Override
        public void visitNode(@Nonnull AstNode node) {
            //选择节点语句
            List<AstNode> nodeDescendants = node.getDescendants(CxxGrammarImpl.selectionStatement);
            for (AstNode astNode : nodeDescendants) {
                //判断节点是不是if节点
                if ("if".equals(astNode.getToken().getValue())) {
                    //获取其中的参数
                    List<AstNode> astNodeList = astNode.getDescendants(CxxGrammarImpl.expressionList);
                    for (AstNode expr : astNodeList) {
                        lists.add(expr.getToken().getValue());
                    }

                    if (astNodeList.size() == 0) {
                        List<AstNode> astNodes = astNode.getDescendants(CxxGrammarImpl.relationalExpression);
                        for (AstNode as : astNodes) {
                            List<AstNode> children = as.getChildren();
                            for (AstNode chil : children) {
                                lists.add(chil.getTokenValue());
                            }
                        }
                    }
                }
            }

            List<AstNode> astNodes = node.getDescendants(CxxGrammarImpl.declarationStatement);
            if (astNodes !=null && !astNodes.isEmpty()) {
                for (AstNode astNode : astNodes) {
                    //获取参数
                    AstNode firstDescendant = astNode.getFirstDescendant(CxxGrammarImpl.initializer);
                    if (firstDescendant != null) {
                        List<AstNode> children = firstDescendant.getChildren();
                        for (AstNode chil : children) {
                            //判断参数类型
                            if ("STRING".equals(chil.getName())) {
                                String result = chil.getTokenValue().replace("\"", "");
                                if (isPath(result)) {
                                    //判断是不是windows或linux的绝对路径
                                    if (!isWindowsAbsolutePath(result) && !isLinuxAbsolutePath(result)) {
                                        reportIssue(chil, "使用关键资源时指定资源所在的路径");
                                    }
                                }
                            }
                        }
                    }
                }
            }
            List<AstNode> descendants = node.getDescendants(CxxGrammarImpl.expression);
            for (AstNode ast : descendants) {
                AstNode descendant = ast.getFirstDescendant(CxxGrammarImpl.postfixExpression);
                if (descendant != null) {
                    String methodName = descendant.getTokenValue();
                    if ("system".equals(methodName) || "fopen".equals(methodName)) {
                        List<AstNode> childrens = descendant.getChildren();
                        if (childrens != null && !childrens.isEmpty()) {
                            String tokenValue = childrens.get(0).getTokenValue();
                            if (tokenValue != null && !lists.equals(tokenValue)) {
                                reportIssue(descendant, "使用关键资源时指定资源所在的路径");
                            }
                        }
                    }
                }
            }
        }
    }
}
