/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：28所 C++ 信息安全性设计准则
 * 项目描述：用于检查C++源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package org.sonar.cxx.squidbridge; // cxx: in use

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import com.google.common.base.Throwables;
import com.google.common.collect.Lists;
import com.sonar.cxx.sslr.api.AstNode;
import com.sonar.cxx.sslr.api.Grammar;
import com.sonar.cxx.sslr.api.RecognitionException;
import com.sonar.cxx.sslr.impl.Parser;
import com.sonar.cxx.sslr.impl.ast.AstWalker;
import java.io.File;
import java.io.InterruptedIOException;
import java.util.Collection;
import java.util.List;
import javax.annotation.Nullable;
import org.sonar.api.batch.fs.InputFile;
import org.sonar.api.utils.log.Logger;
import org.sonar.api.utils.log.Loggers;
import org.sonar.cxx.squidbridge.api.AnalysisException;
import org.sonar.cxx.squidbridge.api.SourceCodeSearchEngine;
import org.sonar.cxx.squidbridge.api.SourceCodeTreeDecorator;
import org.sonar.cxx.squidbridge.api.SourceProject;
import org.sonar.cxx.squidbridge.indexer.SquidIndex;
import org.sonar.cxx.squidbridge.measures.MetricDef;

public class AstScanner<G extends Grammar> {

    private static final Logger LOG = Loggers.get(AstScanner.class);

    private final List<SquidAstVisitor<G>> visitors;
    private final Parser<G> parser;
    private final SquidAstVisitorContextImpl<G> context;

    private final SquidIndex indexer = new SquidIndex();
    private final MetricDef[] metrics;
    private final MetricDef filesMetric;

    protected AstScanner(Builder<G> builder) {
        this.visitors = Lists.newArrayList(builder.visitors);
        this.parser = builder.baseParser;
        this.context = builder.context;

        this.context.setGrammar(parser.getGrammar());
        this.context.getProject().setSourceCodeIndexer(indexer);
        this.context.setCommentAnalyser(builder.commentAnalyser);
        this.metrics = builder.metrics;
        this.filesMetric = builder.filesMetric;
        indexer.index(context.getProject());
    }

    public SourceCodeSearchEngine getIndex() {
        return indexer;
    }

    public void scanFile(File file) {
        scanFiles(java.util.List.of(file));
    }

    public void scanInputFile(InputFile inputFile) {
        scanInputFiles(java.util.List.of(inputFile));
    }

    public void scanFiles(Collection<File> files) {
        initVisitors();

        var astWalker = new AstWalker(visitors);

        for (var file : files) {
            checkCancel();
            context.setFile(file, filesMetric);

            Exception parseException = null;
            AstNode ast = null;
            try {
                try {
                    ast = parser.parse(file);
                } catch (Exception e) {
                    parseException = handleParseException(file, e);
                }
                walkAndVisit(astWalker, ast, parseException);
            } catch (Throwable e) {
                throw new AnalysisException("Unable to parse file: " + file.getAbsolutePath(), e);
            }
        }

        destroyVisitors();
        decorateSquidTree();
    }

    public void scanInputFiles(Iterable<InputFile> inputFiles) {
        initVisitors();

        var astWalker = new AstWalker(visitors);

        for (var inputFile : inputFiles) {
            var file = new File(inputFile.uri().getPath());
            checkCancel();
            context.setInputFile(inputFile, filesMetric);

            Exception parseException = null;
            AstNode ast = null;
            try {
                try {
                    ast = parser.parse(inputFile.contents());
                } catch (Exception e) {
                    parseException = handleParseException(file, e);
                }
                walkAndVisit(astWalker, ast, parseException);
            } catch (Throwable e) {
                LOG.error("Unable to parse file: " + file.getAbsolutePath(), e);
            }
        }

        destroyVisitors();
        decorateSquidTree();
    }

    private static Exception handleParseException(File file, Exception e) {
        checkInterrupted(e);
        if (e instanceof RecognitionException) {
            LOG.error("Unable to parse file: " + file.getAbsolutePath());
            LOG.error(e.getMessage());
        } else {
            LOG.error("Unable to parse file: " + file.getAbsolutePath(), e);
        }
        return e;
    }

    private void walkAndVisit(AstWalker astWalker, AstNode ast, @Nullable Exception parseException) throws Throwable {
        if (parseException == null) {
            astWalker.walkAndVisit(ast);
        } else {
            // process parse error
            for (var visitor : visitors) {
                visitor.visitFile(ast);
            }
            for (var visitor : visitors) {
                if (visitor instanceof AstScannerExceptionHandler) {
                    if (parseException instanceof RecognitionException) {
                        ((AstScannerExceptionHandler) visitor)
                                .processRecognitionException((RecognitionException) parseException);
                    } else {
                        ((AstScannerExceptionHandler) visitor).processException(parseException);
                    }
                }
            }
            for (var visitor : visitors) {
                visitor.leaveFile(ast);
            }
        }
        context.popTillSourceProject();
    }

    private void initVisitors() {
        for (var visitor : visitors) {
            visitor.init();
        }
    }

    private void destroyVisitors() {
        for (var visitor : visitors) {
            visitor.destroy();
        }
    }

    /**
     * Checks if the root cause of the thread is related to an interrupt.
     * Note that when such an exception is thrown, the interrupt flag is reset.
     */
    private static void checkInterrupted(Exception e) {
        Throwable cause = Throwables.getRootCause(e);
        if (cause instanceof InterruptedException || cause instanceof InterruptedIOException) {
            throw new AnalysisException("Analysis cancelled", e);
        }
    }

    private static void checkCancel() {
        if (Thread.interrupted()) {
            throw new AnalysisException("Analysis cancelled");
        }
    }

    protected void decorateSquidTree() {
        if (metrics != null && metrics.length > 0) {
            SourceProject project = context.getProject();
            var decorator = new SourceCodeTreeDecorator(project);
            decorator.decorateWith(metrics);
        }
    }

    public static <G extends Grammar> Builder<G> builder(SquidAstVisitorContextImpl<G> context) {
        return new Builder<>(context);
    }

    public static class Builder<G extends Grammar> {

        private Parser<G> baseParser;
        private final List<SquidAstVisitor<G>> visitors = Lists.newArrayList();
        private final SquidAstVisitorContextImpl<G> context;
        private CommentAnalyser commentAnalyser;
        private MetricDef[] metrics;
        private MetricDef filesMetric;

        public Builder(SquidAstVisitorContextImpl<G> context) {
            checkNotNull(context, "context cannot be null");
            this.context = context;
        }

        public Builder<G> setBaseParser(Parser<G> baseParser) {
            checkNotNull(baseParser, "baseParser cannot be null");
            this.baseParser = baseParser;
            return this;
        }

        public Builder<G> setCommentAnalyser(CommentAnalyser commentAnalyser) {
            checkNotNull(commentAnalyser, "commentAnalyser cannot be null");
            this.commentAnalyser = commentAnalyser;
            return this;
        }

        public Builder<G> withSquidAstVisitor(SquidAstVisitor<G> visitor) {
            checkNotNull(visitor, "visitor cannot be null");
            if(visitor.getContext() == null){
                visitor.setContext(context);
                visitors.add(visitor);
            }
            return this;
        }

        public Builder<G> withMetrics(MetricDef... metrics) {
            for (var metric : metrics) {
                checkNotNull(metric, "metrics cannot be null");
            }
            this.metrics = metrics.clone();
            return this;
        }

        public Builder<G> setFilesMetric(MetricDef filesMetric) {
            checkNotNull(filesMetric, "filesMetric cannot be null");
            this.filesMetric = filesMetric;
            return this;
        }

        public AstScanner<G> build() {
            checkState(baseParser != null, "baseParser must be set");
            checkState(commentAnalyser != null, "commentAnalyser must be set");
            checkState(filesMetric != null, "filesMetric must be set");
            return new AstScanner<>(this);
        }
    }

}
