/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：信息安全性设计准则检查插件
 * 项目描述：用于检查源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */

package com.keyware.sonar.java.rules.checkers;

import org.sonar.check.Rule;
import org.sonar.plugins.java.api.IssuableSubscriptionVisitor;
import org.sonar.plugins.java.api.tree.*;

import java.util.Collections;
import java.util.List;

/**
 * 应设置X-Frame-Options的值为deny
 * 1.获取所有方法调用节点
 * 2.判断方法中的参数是否否和要求
 *
 * @author RenFengJiang
 * @date 2024/1/22
 */
@Rule(key = "OptionsVerifyChecker")
public class OptionsVerifyChecker extends IssuableSubscriptionVisitor {


    @Override
    public List<Tree.Kind> nodesToVisit() {
        return Collections.singletonList(Tree.Kind.METHOD);
    }

    @Override
    public void visitNode(Tree tree) {
        MethodTree node = (MethodTree) tree;
        List<VariableTree> parameters = node.parameters();
        //盘带是否是文件上传类
        boolean booleanRestorn = parameters.stream().anyMatch(type -> "HttpServletResponse".equals(type.type().toString()));
        if (booleanRestorn) {
            MethodCall methodCall = new MethodCall();
            node.block().accept(methodCall);
            if (methodCall.getBooleanInner()) {
                context.reportIssue(this, node.simpleName(), "应设置X-Frame-Options的值为deny");
            }
        }
        DeclarationVerdict declarationVerdict = new DeclarationVerdict(node,this);
        node.block().accept(declarationVerdict);
    }

    private class MethodCall extends BaseTreeVisitor {
        private boolean booleanInner = true;

        public boolean getBooleanInner() {
            return booleanInner;
        }

        public void visitMethodInvocation(MethodInvocationTree tree) {
            if (booleanInner) {
                ExpressionTree expressionTree = tree.methodSelect();
                if (expressionTree.is(Tree.Kind.MEMBER_SELECT)) {
                    MemberSelectExpressionTree selectExpressionTree = (MemberSelectExpressionTree) expressionTree;
                    if ("addHeader".equals(selectExpressionTree.identifier().name()) || "setHeader".equals(selectExpressionTree.identifier().name())) {
                        Arguments arguments = tree.arguments();
                        boolean one = false;
                        boolean two = false;
                        for (ExpressionTree argument : arguments) {
                            if (argument.is(Tree.Kind.STRING_LITERAL)) {
                                String literalValue = ((LiteralTree) argument).value();
                                if ("\"X-Frame-Options\"".equals(literalValue)) {
                                    one = true;
                                } else if ("\"DENY\"".equals(literalValue) ||"\"deny\"".equals(literalValue)) {
                                    two = true;
                                }
                            }
                        }
                        if (one && two) {
                            booleanInner = false;
                        }
                    }
                }
            }
        }
    }

    public class DeclarationVerdict extends BaseTreeVisitor {

        private MethodTree node;
        private OptionsVerifyChecker checker;
        public DeclarationVerdict( MethodTree node,OptionsVerifyChecker checker){
            this.node = node;
            this.checker = checker;
        }


        @Override
        public void visitVariable(VariableTree tree) {
            IdentifierTree identifierTree = tree.simpleName();
            TypeTree type = tree.type();
            if(type != null && type instanceof IdentifierTree){
                IdentifierTree fierTree = (IdentifierTree) type;
                if("HttpServletResponse".equals(fierTree.name())){
                    MethodCall methodCall = new MethodCall();
                    node.block().accept(methodCall);
                    if (methodCall.getBooleanInner()) {
                        context.reportIssue(checker, identifierTree, "应设置X-Frame-Options的值为deny");
                    }
                }
            }
        }
    }
}
