/*
 * Copyright (c) 2023 - 2024. KeyWare.Co.Ltd All rights reserved.
 * 项目名称：信息安全性设计准则检查插件
 * 项目描述：用于检查源代码的安全性设计准则的Sonarqube插件
 * 版权说明：本软件属北京关键科技股份有限公司所有，在未获得北京关键科技股份有限公司正式授权情况下，任何企业和个人，不能获取、阅读、安装、传播本软件涉及的任何受知识产权保护的内容。
 */
package com.keyware.sonar.java.rules.checkers;

import org.sonar.check.Rule;
import org.sonar.plugins.java.api.IssuableSubscriptionVisitor;
import org.sonar.plugins.java.api.tree.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * 使用RSA最优加密填充
 *
 * @author RenFengJiang
 * @date 2024/1/11
 */
@Rule(key = "RSAEncryptionChecker")
public class RSAEncryptionChecker extends IssuableSubscriptionVisitor {

    private List nameLists = new ArrayList();

    @Override
    public List<Tree.Kind> nodesToVisit() {
        /**
         * Tree.Kind.METHOD：方法节点
         * Tree.Kind.BLOCK：方法的代码块节点
         * Tree.Kind.METHOD_INVOCATION： 方法的调用节点
         */
        return Arrays.asList(Tree.Kind.METHOD_INVOCATION, Tree.Kind.CLASS);
    }

    @Override
    public void visitNode(Tree tree) {

        if (tree.is(Tree.Kind.CLASS)) {
            MethodBOdyVisitor methodBOdyVisitor = new MethodBOdyVisitor();
            tree.accept(methodBOdyVisitor);
            nameLists = methodBOdyVisitor.getNameLists();
        } else if (tree.is(Tree.Kind.METHOD_INVOCATION)) {

            MethodInvocationTree methodInvocationTree = (MethodInvocationTree) tree;
            ExpressionTree expressionTree = methodInvocationTree.methodSelect();
            if (expressionTree != null) {
                // 获取到方法调用
                if (expressionTree instanceof MemberSelectExpressionTree) {
                    MemberSelectExpressionTree memberSelectExpressionTree = (MemberSelectExpressionTree) expressionTree;
                    // 判断是否符合标准
                    String a = memberSelectExpressionTree.expression().toString();
                    String b = memberSelectExpressionTree.identifier().name();
                    if ("Cipher".equals(memberSelectExpressionTree.expression().toString()) && "getInstance".equals(memberSelectExpressionTree.identifier().name())) {
                        // 获取参数列表
                        List<ExpressionTree> arguments = methodInvocationTree.arguments();
                        for (ExpressionTree argument : arguments) {
                            if (argument.is(Tree.Kind.STRING_LITERAL)) {
                                LiteralTree literalTree = (LiteralTree) argument;
                                String c = ((LiteralTree) argument).token().text();
                                // 对参数进行判断判断是否符合要求
                                if (!literalTree.token().text().startsWith("\"RSA")) {
                                    context.reportIssue(this, argument, "使用RSA最优加密填充");
                                }
                            } else if (!nameLists.equals(argument.toString())) {
                                context.reportIssue(this, argument, "使用RSA最优加密填充");
                            }
                        }
                    }
                }
            }
        }

    }

    static class MethodBOdyVisitor extends BaseTreeVisitor {

        private List nameLists = new ArrayList();

        public MethodBOdyVisitor() {

        }

        public List getNameLists() {
            return nameLists;
        }

        @Override
        public void visitVariable(VariableTree tree) {
            IdentifierTree identifierTree = tree.simpleName();
            ExpressionTree initializer = tree.initializer();
            if (identifierTree != null && initializer != null && initializer instanceof LiteralTree) {
                LiteralTree literalTree = (LiteralTree) initializer;
                if (literalTree.value().startsWith("\"RSA")) {
                    nameLists.add(identifierTree.name());
                }
            }


        }
    }
}
